/* Copyright 2019 Benjamin Worpitz
 *
 * This file is part of alpaka.
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */

#pragma once

#ifndef ALPAKA_ACC_CPU_B_SEQ_T_OMP2_ENABLED
#    define ALPAKA_ACC_CPU_B_SEQ_T_OMP2_ENABLED
#endif
