
//              Copyright Catch2 Authors
// Distributed under the Boost Software License, Version 1.0.
//   (See accompanying file LICENSE.txt or copy at
//        https://www.boost.org/LICENSE_1_0.txt)

// SPDX-License-Identifier: BSL-1.0

#include <catch2/generators/catch_generators_random.hpp>

#include <catch2/internal/catch_context.hpp>

std::uint32_t Catch::Generators::Detail::getSeed() { return sharedRng()(); }
